/*
  LinEAK - Linux support for Easy Access and Internet Keyboards      
  Copyright (C) 2001,2002 Mark Smulders <Mark@PIRnet.nl>             
  http://lineak.sourceforge.net

  [NOTE]
    The X11/Xkb related functions in this file (initializeX, GetDisplay, commitXKBchanges)
	are based on the work of Anthony Y P Wong <ypwong@ypwong.org>
    I owe him a great deal for inspiring me with HOTKEYS.
    Thanks Anthony!
	
  This program is free software; you can redistribute it and/or modify  
  it under the terms of the GNU General Public License as published by  
  the Free Software Foundation; either version 2, or (at your option)   
  any later version.                                                    
                                                                        
  This program is distributed in the hope that it will be useful,       
  but WITHOUT ANY WARRANTY; without even the implied warranty of        
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
  GNU General Public License for more details.                          
                                                                        
  You should have received a copy of the GNU General Public License     
  along with this program; if not, write to the                         
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,          
  Boston, MA 02111-1307, USA.                                           
*/

/* settings from configure */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/* system header files */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#ifdef HAVE_GETOPT_H
#  define HAVE_GETOPT_LONG
#  include <getopt.h>
#endif
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/XKBfile.h>
#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
extern char *getenv();
#endif
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h> /* linux specific?!! */

extern int mkdir(); /* in linux/fs.h ... other platforms!? */

#if defined (__FreeBSD__)
# include <machine/soundcard.h>
#else
# if defined (__NetBSD__) || defined (__OpenBSD__)
#  include <soundcard.h>          /* OSS emulation */
#  undef ioctl
# else
/* BSDI, Linux, Solaris */
#  include <sys/soundcard.h>
# endif
#endif

/* lineakd header file */
#include "lineakd.h"

/**********************************************************************************************/

/* main function */
int main(int argc, char *argv[]) {
  XkbEvent xev;
  
  /* read the configuration and keyboard definition file */
  parseconffile();
  parsedeffile();
  /* parse command line arguments */
  if ( !parseargs(argc,argv) )
	cleanexit(3);
  
  /* check if a pid file exists */
  if (pid_file_exists())
	cleanexit(3);
  
  /* everything went according to plan thisfar... some signal-handlers for cleaner exiting */
  signal(SIGTERM, signalexit);
  signal(SIGINT, signalexit);
  /* and one so we won't have to wait() for child processes ;) */
  signal(SIGCLD, SIG_IGN);
  /* and for a rehash when we catch SIGHUP */
  signal(SIGHUP,signalHUP);
  
  /* init the cdrom (disable auto eject) */
  if (strcmp(cdromdev,"")!=0) { initCDROM (); }

  /* init sound (store master volume of MIXERDEV) */
  initSound();
  
  /* build our EAKeyboard with the data from deffile and conffile */
  initEAK();
  
  /* init X/Xkb with our EAK keycodes/keysyms */
  initializeX();
  
  /* do we daemonize now? do we? do we? ...ahhh, please? ;-) */
  if (runasdaemon)
	daemon(0,0); // stdin, stdout and stderr to /dev/null when daemon...
  
  /* write the pid file (after daemon check, or the pid won't be right) */
  create_pid_file();
  
  /* alright, we're going to loop forever now. only signals can interrupt us. */
  while(1) {
	/* wait for next event */
	XNextEvent(ourdisplay, &xev.core);
	/* should we respond to this event? */
    if ( xev.type == xkbEventCode+XkbEventCode && xev.any.xkb_type == XkbActionMessage ) {
      if (verbosemode) printf("Keycode %d pressed\n", xev.message.keycode);
	  /* find the corresponding key, and execute it's command */
	  execEAcommand(xev.message.keycode);
	}
  }
  
  /* that's all folks.. */
  cleanup(); // just for fun.. we don't ever get here ;-)
  return 0;
}

/**********************************************************************************************/

/* parse .conf file */
void parseconffile(void) {
  FILE *fp = NULL;
  char *homedir = NULL;

  /* SET full path of config file ($HOME/.lineak/CONFIGFILE) */
  homedir = getenv("HOME");
  conffile = malloc((strlen(homedir) + 1 + 7 + 1 + strlen(CONFIGFILE) + 1)*sizeof(char));
  strcpy(conffile, homedir);
  strcat(conffile, "/");
  strcat(conffile, ".lineak");
  strcat(conffile, "/");
  strcat(conffile, CONFIGFILE);
  strcat(conffile, "\0");

  /* does the file exist? */
  if ((fp = fopen(conffile, "r")) == NULL) {
	/* file does not exist, create standard one */
    printf("\nNOTE: configuration file not found,\n");
	create_new_conf(homedir, kbtype);
  } else {
    fclose(fp);
  }
  /* now that we're sure it exists, parse the contents */
  if ((cfgParse(conffile, confdata, CFG_SIMPLE)) == -1) {
	/* error parsing file */
	printf("FATAL ERROR occurred while reading configuration data!\n");
	cleanexit(3);
  }
}

/* create new .conf file in homedir for keyboard 'type' */
void create_new_conf(char *homedir, char *type) {
  FILE *fp;
  char *lineakdir, *EAKname;
  int i=0, EAKnr;
  cfgList *l;

  /* get homedir */
  if (homedir == NULL)
    homedir = getenv("HOME");	
  /* check if $HOME/.lineak/ directory exists, if not create it */
  lineakdir = malloc( (strlen(homedir) + 1 + 7 + 1) * sizeof(char));
  strcpy(lineakdir, homedir);
  strcat(lineakdir, "/.lineak\0");
  if (opendir(lineakdir) == NULL) {
    if (mkdir(lineakdir, 0755) == -1) {
      printf("FATAL ERROR: unable to create directory %s !\n",lineakdir);
	  cleanexit(3);
    } 
  }
  
  /* open the file for write */
  if ((fp = fopen(conffile, "w")) != NULL) {
    /* no CD-ROM device specified? */
    if ((cdromdev==NULL) || (cdromdev[0]=='\0'))
	  printf("Warning: no CD-ROM device specified!\n");

	if (type != NULL) { /* create kb specific .conf */
      /* do we have a valid keyboard type? */
      if (cfgSectionNameToNumber(type) == -1) {
	    printf("\nInvalid keyboard type: %s\nTo find out the supported keyboard types, use: '%s -l'\n\n",type,"lineakd");
	    cleanexit(3);
      }
	  printf("\nNOTE: Creating %s for keyboard type: %s\n",conffile,type);
      EAKnr = cfgSectionNameToNumber(type);
      EAKname = longkbname[EAKnr];
      fprintf(fp,"%s# LinEAK Configuration file for %s\n\n# Global settings\n"
                 "KeyboardType\t\t= %s\n"
                 "CdromDevice\t\t= %s\n"
                 "\n# Specific keys of your keyboard\n", CONF_HEADER, EAKname, type, cdromdev==NULL?"":cdromdev);
	  for (l=keylist[EAKnr]; l!=NULL; l=l->next)
        fprintf(fp,"%s%s= \n",l->str,(strlen(l->str)>7)?"\t":"\t\t");
	} else { /* create generic .conf containing all keys */
	  printf("\nNOTE: Creating %s using generic default values.\n      To create a keyboard-specific file, please run 'lineakd -c TYPE'\n",conffile);
      fprintf(fp,"%s# LinEAK Configuration file. This generic .conf file contains ALL keys.\n#  To create a keyboard-specific file, please run 'lineakd -c TYPE'\n\n# Global settings\n"
                 "KeyboardType\t\t= CEAK8\n"
                 "CdromDevice\t\t= %s\n"
                 "\n# ALL key definitions\n", CONF_HEADER, cdromdev==NULL?"":cdromdev);
      while (confdata[i].parameterName != NULL) {
	    fprintf(fp,"%s%s= \n",confdata[i].parameterName,(strlen(confdata[i].parameterName)>7)?"\t":"\t\t");
	    i++;
 	  }
	}
	fprintf(fp,"\n# end lineakd.conf\n");
    fclose(fp);	
	printf("NOTE: Now please edit the file and add commands to it,\n      or use lineakconfig :)\n\n");
  } else {
    printf("FATAL ERROR: unable to write configuration file %s !\n",conffile);
	cleanexit(3);
  }
}

/* parse keyboard definition file */
void parsedeffile(void) {
  FILE *fp = NULL;

  /* SET full path of definition file ($sysconfdir/lineakkb.def) */
  deffile = malloc((strlen(CONFDIR) + 1 + strlen(DEFINITIONFILE) + 1)*sizeof(char));
  strcpy(deffile, CONFDIR);
  strcat(deffile, "/");
  strcat(deffile, DEFINITIONFILE);
  strcat(deffile, "\0");

  /* does the file exist? */
  if ((fp = fopen(deffile, "r")) != NULL) {
	fclose(fp);
	/* it exists, parse the contents */
	if ((nrkeyboards = cfgParse(deffile, defdata, CFG_INI)) == -1) {
	  /* error parsing file */
	  printf("FATAL ERROR occurred while reading definition data!\n");
  	  cleanexit(3);
	}
  } else {
	/* file does not exist */
    printf("FATAL ERROR: definition file %s not found!\n",deffile);
	cleanexit(3);
  }
}

/* check if the PID file exists */
int pid_file_exists(void) {
  FILE *fp;
  char *home, *pidfile;
  home = getenv("HOME");
  pidfile = malloc ( ( strlen(home) + 1 + 7 + 1 + strlen(PIDFILE) + 1) * sizeof(char));
  sprintf(pidfile, "%s/.lineak/%s", home, PIDFILE);
  if ((fp = fopen(pidfile, "r"))!=NULL) {
	printf(" *** lineakd already running!\n *** if not then delete %s\n", pidfile);
	fclose (fp);
	return 1;
  } else
	return 0;
}

/* create the PID file */
void create_pid_file(void) {
  pid_t mypid;
  char *home, *pidfile;
  FILE *fp;
  mypid = getpid();
  home = getenv("HOME");
  pidfile = malloc ( ( strlen(home) + 1 + 7 + 1 + strlen(PIDFILE) + 1) * sizeof(char));
  sprintf(pidfile, "%s/.lineak/%s", home, PIDFILE);
  if ((fp = fopen(pidfile, "w"))!=NULL) { 
	fprintf(fp, "%d", mypid);
	fclose (fp);
	free(home);
	free(pidfile);
  } else {
	printf("FATAL ERROR writing PID file %s\n", pidfile);
	free(home);
	free(pidfile);
	cleanexit(1);
  }
}

/* for deleting the PID file */
void delete_pid_file(void) {
  char *home, *rmpidfile;
  home = getenv("HOME");
  rmpidfile = malloc ( ( 3 + strlen(home) + 1 + 7 + 1 + strlen(PIDFILE) + 1) * sizeof(char));
  sprintf(rmpidfile, "rm %s/.lineak/%s", home, PIDFILE);
  system(rmpidfile);
  free(home);
  free(rmpidfile);
}

/**********************************************************************************************/

/* show supported keyboards */
void showkeyboards (void) {
  int i;
  printf("\nLinEAK v" VERSION " -- supported keyboards:\n\n");
  printf(" [TYPE]\t\t[Full name]\n\n");
  for (i=0; i<nrkeyboards; i++) {
	printf(" %s%s(%s)\n",cfgSectionNumberToName(i),strlen(cfgSectionNumberToName(i))>7?"\t":"\t\t",longkbname[i]);
  }
  printf("\n");
}

/* init the EAKeyboard */
int initEAK (void) {
  cfgList *l = NULL;
  EAkey *tmp = NULL;
  int keycount = 0;
  myEAK.EAKnr = cfgSectionNameToNumber(kbtype);
  myEAK.EAKtype = kbtype;
  myEAK.EAKname = longkbname[myEAK.EAKnr];
  if (verbosemode) printf("Configured your keyboard as \"%s\" (nr: %d  name: %s)\n",myEAK.EAKtype,myEAK.EAKnr,myEAK.EAKname);
  if (keylist[myEAK.EAKnr] != NULL) {
	myEAK.EAKeylist = tmp = malloc(sizeof(EAkey));
	for (l=keylist[myEAK.EAKnr]; l!=NULL; l=l->next) {
      keycount++;
	  tmp->EAkeyname = l->str;
	  if (fetchVarFromCfgFile(deffile, tmp->EAkeyname, &(tmp->EAkeycode), CFG_INT, CFG_INI, 0, kbtype) == -1) {
		printf("FATAL ERROR while fetching keycode for key '%s' from definition file!\n",tmp->EAkeyname);
	    cleanexit(1);
	  }
	  tmp->EAkeysym = 0x1008FF01 + keycount;
	  tmp->EAcommand = NULL;
	  if (fetchVarFromCfgFile(conffile, tmp->EAkeyname, &(tmp->EAcommand), CFG_STRING, CFG_SIMPLE, 0, NULL) == -1) {
		printf("FATAL ERROR while fetching command for key '%s' from configuration file!\n",tmp->EAkeyname);
	    cleanexit(1);
	  }
	  if (verbosemode) printf("Configured key \"%s\" with keycode %d, keysym 0x%x and command: %s\n",
		                       tmp->EAkeyname,tmp->EAkeycode,(unsigned int)tmp->EAkeysym,
	                           strcmp(tmp->EAcommand,"")==0?"(none)":tmp->EAcommand);
	  if (l->next != NULL) {
		tmp->next = malloc(sizeof(EAkey));
		tmp = tmp->next;
	  } else {
		tmp->next = NULL;
	  }
	}
  } else {
	myEAK.EAKeylist = NULL;
  }
  return 0;
}

/* run the proper command of the key with keycode 'keycode' */
void execEAcommand(int keycode) {
  EAkey *theKey = NULL;
  char *theCmd = NULL;
  /* lookup the key in our EAKeylist */
  if ((theKey = lookupEAkey(keycode)) == NULL)
	/* oops, not found... how did THAT happen? */
	printf("Key with keycode %d not found!!\n",keycode);
  else {
	if (verbosemode) printf("... that's the \"%s\" key.\n",theKey->EAkeyname);
	/* check if the key has to run a SPECIAL command */
	if (strcmp(theKey->EAcommand,EAK_EJECT)==0) { /* Eject */
      if (strcmp(cdromdev,"")!=0) {
		DoEject();
	  }
	} else if (strcmp(theKey->EAcommand,EAK_VOLUP)==0) { /* VolumeUp */
	  AdjustVolume(VOLUP_VALUE);
	} else if (strcmp(theKey->EAcommand,EAK_VOLDOWN)==0) { /* VolumeDown */
	  AdjustVolume(VOLDOWN_VALUE);
	} else if (strcmp(theKey->EAcommand,EAK_MUTE)==0) { /* Mute */
	  AdjustVolume(VOLMUTE_VALUE);
	} else if (strcmp(theKey->EAcommand,EAK_SLEEP)==0) { /* Sleep (not implemented yet) */
      
    } else if (strcmp(theKey->EAcommand,"")==0) { /* no command defined */
	  if (verbosemode) printf("... but it has no command bound to it :(\n");
	} else {
	  /* all specials done, let's go for it... ;) */
	  if (verbosemode) printf("... running \"%s\"\n",theKey->EAcommand);
	  if (!fork()) {
		/* child process that tries to run the command, and then exits */
		theCmd = malloc((strlen(theKey->EAcommand)+3)*sizeof(char));
		strcpy(theCmd,theKey->EAcommand);
		strcat(theCmd," &\0");
		system(theCmd);
		free(theCmd);
		//system(theKey->EAcommand " &");
		cleanexit(3);
	  }
	  /* NOTE: no, we don't wait() for the child. we just ignore the SIGCLD signal */
	}
  }
}

/* find the key with keycode 'keycode' in myEAK.EAKeylist */
EAkey *lookupEAkey(int keycode) {
  EAkey *tmp = NULL;
  for (tmp = myEAK.EAKeylist; tmp!=NULL; tmp=tmp->next)
	if (tmp->EAkeycode == keycode)
	  break;
  return tmp;	  
}

/**********************************************************************************************/

/* init CD-ROM: disable SW eject */
void initCDROM (void) {
  int fp;
  if (verbosemode) printf("CD-ROM init, using \"%s\" as the CD-ROM device\n",cdromdev);
  /* do we have a CD-ROM device configured?  (REDUNDANT extra check) */
  if (strcmp(cdromdev,"")!=0) {
	/* try to open the device .. */
    if ( (fp = open(cdromdev, O_RDONLY | O_NONBLOCK)) != -1 ) {
	  /* tell the drive to diable auto-eject */
	  if ( (ioctl(fp, CDROMEJECT_SW, 0)) == -1 )
        printf("... oops! error during CD-ROM init\n");
	  close(fp);
	} else {
	  printf("... oops! unable to open the CD-ROM device \"%s\" (CD-ROM init)\n",cdromdev);
	}
  } else {
	printf("no CD-ROM device configured! (CD-ROM init)\n");
  }
}

/* close or eject CD-ROM drive */
void DoEject(void) {
  int fp;
  /* do we have a CD-ROM device configured?  (extra check, not really nescessary) */
  if (strcmp(cdromdev,"")!=0) {
	/* try to open the device .. */
    if ( (fp = open(cdromdev, O_RDONLY | O_NONBLOCK)) != -1 ) {
	  if (!cdromstatus) { /* assumed closed */
		/* enable the drives software eject */
		if ( (ioctl(fp, CDROMEJECT_SW, 1)) == -1 )
          printf("... oops! error enabling CD-ROM SW eject\n");
		/* eject the cdrom tray */
		if (verbosemode) printf("... ejecting the CD-ROM tray\n");
        if ( (ioctl(fp, CDROMEJECT)) == -1 )
          printf("... oops! failed to eject the CD-ROM\n");
		else
		  cdromstatus = !cdromstatus;
	  } else { /* assumed open */
        /* disable the drives software eject again.. */
        if ( (ioctl(fp, CDROMEJECT_SW, 0)) == -1 )
          printf("... oops! error disabling CD-ROM SW eject\n");
		/* close the cdrom tray */
		if (verbosemode) printf("... closing the CD-ROM tray\n");
        if ( (ioctl(fp, CDROMCLOSETRAY)) == -1 )
          printf("... oops! failed to close the CD-ROM\n");
		else
		  cdromstatus = !cdromstatus;
	  }
      close(fp);
	} else {
	  printf("... oops! unable to open the CD-ROM device \"%s\"\n",cdromdev);
	}
  } else {
	printf("... oops! no CD-ROM device configured!\n");
  }
}

/**********************************************************************************************/

/* init sound: read and store the mixer master volume */
void initSound(void) {
  int fp;
  if (verbosemode) printf("Sound init, using \"%s\" as the mixer device\n",MIXERDEV);
  if ( (fp = open( MIXERDEV, O_RDWR|O_NONBLOCK )) != -1 ) {
    if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &old_mastervol) != -1) {
	  if (verbosemode) printf("... master volume stored\n");
	} else {
      printf("... oops! unable to read the volume of \"%s\" (sound init)\n", MIXERDEV);
 	}
  } else {
    printf("... oops! unable to open the mixer device \"%s\" (sound init)\n", MIXERDEV);
  }
}

/* adjust master volume or mute it */
void AdjustVolume(int value) {
  int fp, master_vol, left, right;
  /* open the mixer device */
  if ( (fp = open( MIXERDEV, O_RDWR|O_NONBLOCK )) != -1 ) {
	if (value == VOLUP_VALUE || value == VOLDOWN_VALUE) { /* volume up/down */
	  if (ismuted) { if (verbosemode) printf("... but we're muted\n"); close(fp); return; }
	  if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &master_vol) != -1) {
	    /* adjust the master volume */
        left = (master_vol & 0xFF) + value;
        right = ((master_vol >> 8) & 0xFF) + value;
        left = (left>MAX_VOLUME) ? MAX_VOLUME : ((left<0) ? 0 : left);
        right = (right>MAX_VOLUME) ? MAX_VOLUME : ((right<0) ? 0 : right);
        master_vol = left + (right << 8);
        if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &master_vol) != -1) {
          if (verbosemode) printf("... volume adjusted\n");
        } else {
		  printf("... oops! unable to adjust the master volume\n");
        }
      } else {
        printf("... oops! unable to read the volume of \"%s\"", MIXERDEV);
 	  }
	} else if (value == VOLMUTE_VALUE) { 
	  if (!ismuted) { /* mute */
		if ( SOUND_IOCTL(fp, SOUND_MIXER_READ_VOLUME, &old_mastervol) != -1) {
	      if (verbosemode) printf("... old master volume stored\n");
          if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &value) != -1) {
            if (verbosemode) printf("... master volume muted\n");
			ismuted = 1;
          } else {
		    printf("... oops! unable to mute the master volume\n");
          }
	    } else {
          printf("... oops! unable to read the volume of \"%s\"\n", MIXERDEV);
		}
	  } else { /* un-mute */
        if (SOUND_IOCTL(fp, SOUND_MIXER_WRITE_VOLUME, &old_mastervol) != -1) {
          if (verbosemode) printf("... master volume restored\n");
		  ismuted = 0;
        } else {
          printf("... oops! unable to restore the master volume\n");
        }
	  }
	}
    close(fp);
  } else {
    printf("... oops! unable to open the mixer device \"%s\"\n", MIXERDEV);
  }
}

/**********************************************************************************************/

/* these functions provide for a somewhat cleaner exit */
static void cleanexit (int status) {
  cleanup();
  /* special status (3) if PID file may NOT be deleted */
  if (status != 3)
	delete_pid_file();
  exit(status);
}
/* cleanup() only used when really bailing out :) */
static void cleanup (void) {
  /* free the dynamically assigned file names */
  free(conffile);
  free(deffile);
  /* and more importantly, clean up that linked list ;) */
  cleanKeylist();
}

/* destroy the EAkeys linked list */
void cleanKeylist (void) {
  EAkey *tmp = NULL, *tmp2 = NULL;
  tmp = myEAK.EAKeylist;
  while (tmp!=NULL) {
	tmp2 = tmp->next;
	free(tmp->EAcommand);
	free(tmp);
	tmp = tmp2;
  }
}

/* this is our "clean-up-that-mess" signal handler ;-) */
static void signalexit(int sig_num) {
  sigset_t mask_set; /* our signal masking set. */
  sigset_t old_set;  /* the old mask set.   */
  /* mask (block) any further signals while we're inside this signal handler. */
  sigfillset(&mask_set);
  sigprocmask(SIG_SETMASK, &mask_set, &old_set);
  /* now do some cleaning... */
  if (verbosemode) printf("cleaning up...\n");
  cleanup();
  delete_pid_file();
  if (verbosemode) printf("bye!\n");
  /* .. restore the old signal mask .. */
  sigprocmask(SIG_SETMASK, &old_set, NULL);
  /* and finally exit.. */
  exit(1);
}

/* HUP signal handler. reload .conf .def files and re-init program */
static void signalHUP(int sig_num) {
  sigset_t mask_set; /* our signal masking set. */
  sigset_t old_set;  /* the old mask set.   */
  /* mask (block) any further signals while we're inside this signal handler. */
  sigfillset(&mask_set);
  sigprocmask(SIG_SETMASK, &mask_set, &old_set);  
  printf("*** HUP signal caught. rehashing...\n");
  /* destroy EAkeyboard keylist */
  if (verbosemode) printf("*** cleaning up old keylist\n");
  cleanKeylist();
  /* reload .conf and .def file */
  if (verbosemode) printf("*** reloading .conf file \"%s\"\n",conffile);
  parseconffile();
  if (verbosemode) printf("*** reloading .def file \"%s\"\n",deffile);
  parsedeffile();
  /* re-init the EAKeyboard */
  if (verbosemode) printf("*** re-initializing the EAKeyboard\n");
  initEAK();
  /* re-init X/Xkb with our EAK keycodes/keysyms */
  if (verbosemode) printf("*** reconfiguring Xkb keymap with new keycodes/keysyms\n");
  initializeX();
  /* .. restore the old signal mask .. */
  if (verbosemode) printf("*** rehash done.\n");
  sigprocmask(SIG_SETMASK, &old_set, NULL);
  return;
}

/**********************************************************************************************/

/* X error handler. When a BadValue error occurs during XkbSetMap, we want to retry */
static int EAKXhandler(Display* theDisplay, XErrorEvent* xev) {
    // BadValue Error   (these numbers might not be platform or version independant?!!!):
    // xev->request_code = 150 (XKEYBOARD)
    // xev->minor_code   = 9   (XkbSetMap)
  if (nrretries < MAXRETRIES) {
	nrretries++;
    if (xev->request_code==150 && xev->minor_code==9)
      caughtXerror = 1;
  } else {
	/* max nr retries failed, quit to prevent race condition */
    printf("*** Xlib error caught, %d retries failed, exiting...\n",MAXRETRIES);
	cleanexit(1);
  }
  return 0;
}

/* dummy for SetAfterFunction */
static int dummyAfterFunc() { return 0; }

/* initialize X (xkb) so it understands the extra keys */
void initializeX(void) {
  XkbMessageAction xma;
  XkbMapChangesRec mapChangeRec;
  int types[1];
  EAkey *thisKey = NULL;

  ourdisplay = GetDisplay(displayname, &xkbOpcode, &xkbEventCode);
  if (!ourdisplay)
    cleanexit(1);

  /* catch Xlib error to retry */
  XSetAfterFunction(ourdisplay, dummyAfterFunc);
  XSetErrorHandler(EAKXhandler);
  
  /* construct the message action struct */
  xma.type = XkbSA_ActionMessage;
  xma.flags = XkbSA_MessageOnPress;
  strcpy(xma.message," ");

  /* add our new keysyms to the key codes of our display */
  for (thisKey = myEAK.EAKeylist; thisKey!=NULL;) {

    /* get the xkb map */
    xkb = XkbGetMap(ourdisplay, XkbAllMapComponentsMask, XkbUseCoreKbd);
    if (!xkb) {
      printf("FATAL ERROR: XkbGetMap failed\n");
  	  cleanexit(1);
    }

	/* check if our new keysym is valid */
	if (!XkbKeycodeInRange(xkb, thisKey->EAkeycode)) {
      printf("The keycode %d cannot be used, as it's not between the min(%d) and max(%d) keycode of your keyboard.\n"
             "Please increase the 'maximum' value in /usr/X11R6/lib/X11/xkb/keycodes/xfree86, then restart X.\n",
             thisKey->EAkeycode, xkb-> min_key_code, xkb->max_key_code);
      continue;
    }	
    /* assign a group to the key code */
	types[0] = XkbOneLevelIndex;
    if ( XkbChangeTypesOfKey(xkb, thisKey->EAkeycode, 1, XkbGroup1Mask, types, NULL) != Success ) {
      printf("FATAL ERROR: XkbChangeTypesOfKey failed");
	  cleanexit(1);
    }
    /* change the keysym */
    if ( XkbResizeKeySyms( xkb, thisKey->EAkeycode, 1 ) == NULL ) {
      printf("FATAL ERROR: resize keysym failed\n");
	  cleanexit(1);
    }
	/* assign our new keysym to the keycode */
    *XkbKeySymsPtr(xkb,thisKey->EAkeycode) = thisKey->EAkeysym;
    /* add one key action to it */
    if ( XkbResizeKeyActions(xkb,thisKey->EAkeycode,1) == NULL ) {
      printf("FATAL ERROR: resize key action failed\n");
	  cleanexit(1);
    }
    commitXKBChanges(thisKey->EAkeycode);
    commitXKBChanges(thisKey->EAkeycode);
	  /* Anthony was right... this really needs to be called twice :)
	     When called only once, and the program is launched when logging into an X session,
	     one of every two keys is not functional ... (still weird thought that when launched
	     from an xterm, calling it only once _does_ work..)
	  */
    /* assign the message action to the key code */
    (&(xkb->server->acts[ xkb->server->key_acts[thisKey->EAkeycode] ]))[0] = (XkbAction) xma;
	/* commit the change back to the server. */
	bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeyActionsMask;
    mapChangeRec.first_key_act = thisKey->EAkeycode;
    mapChangeRec.num_key_acts = 1;
    if ( XkbChangeMap(ourdisplay, xkb, &mapChangeRec) ) {
      if (verbosemode) printf("Xkb map change done for keycode %d\n",thisKey->EAkeycode);
    } else {
      printf("FATAL ERROR: map changed failed\n"); 
	  cleanexit(1);
    }
	if (!caughtXerror)
	  thisKey=thisKey->next;
	else {
	  if (verbosemode) printf("*** caught X error, retrying for keycode %d...\n",thisKey->EAkeycode);
	  caughtXerror = 0;
	}
  }

  /* Select the ActionMessage event under any circumstances */
  if ( !XkbSelectEvents(ourdisplay, XkbUseCoreKbd, XkbActionMessageMask, XkbActionMessageMask)) {
    printf("FATAL ERROR: Couldn't select desired XKB events\n");
    cleanexit(1);
  }
}  

/**********************************************************************************************/

/* Get a hold on the display */
Display *GetDisplay(char* dpyName, int* opcodeRtrn, int* evBaseRtrn) {
  int mjr,mnr,error;
  Display *dpy;

  mjr = XkbMajorVersion;
  mnr = XkbMinorVersion;
  dpy = XkbOpenDisplay(dpyName,evBaseRtrn,NULL,&mjr,&mnr,&error);
  if (dpy == NULL) {
    switch (error) {
            case XkbOD_BadLibraryVersion:
                printf("%s was compiled with XKB version %d.%02d\n",
                        programname,XkbMajorVersion,XkbMinorVersion);
                printf("your X library supports incompatible version %d.%02d\n",
                        mjr,mnr);
                break;
            case XkbOD_ConnectionRefused:
                printf("Cannot open display \"%s\"\n",dpyName);
                break;
            case XkbOD_NonXkbServer:
                printf("XKB extension not present on %s\n",dpyName);
                break;
            case XkbOD_BadServerVersion:
                printf("%s was compiled with XKB version %d.%02d\n",
                        programname,XkbMajorVersion,XkbMinorVersion);
                printf("Server %s uses incompatible version %d.%02d\n",
                        dpyName,mjr,mnr);
                break;
            default:
                printf("Unknown error %d from XkbOpenDisplay\n",error);
      }
    } else {
      if (opcodeRtrn)
        XkbQueryExtension(dpy,opcodeRtrn,evBaseRtrn,NULL,&mjr,&mnr);
    }
    return dpy;
}

/**********************************************************************************************/

void commitXKBChanges(int tcode) {
    XkbMapChangesRec    mapChangeRec;

    /* Commit the change back to the server */
    bzero(&mapChangeRec, sizeof(mapChangeRec));
    mapChangeRec.changed = XkbKeySymsMask | XkbKeyTypesMask;
    mapChangeRec.first_key_sym = tcode;
    mapChangeRec.num_key_syms = 1;
    //        mapChangeRec.first_key_act = tcode;
    //        mapChangeRec.num_key_acts = 1;
    mapChangeRec.first_type = 0;
    mapChangeRec.num_types = xkb->map->num_types;
    if ( XkbChangeMap(ourdisplay, xkb, &mapChangeRec) ) {
      //printf("map change done: %d\n",tcode);
    } else {
      printf("FATAL ERROR: map changed failed\n");
	  cleanexit(1);
    }
}

/**********************************************************************************************/

/* show usage information on stdout */
void showusage(int argc, char *argv[]) {
  printf("\nLinEAK v" VERSION " -- Linux support for Easy Access and Internet Keyboards");
  printf("\n");
  printf("Usage: %s [options...]\n", argv[0]);
  printf("Valid options are:\n");
#ifdef HAVE_GETOPT_LONG
  printf("    -l,        --kbd-list           Show all supported keyboard types\n");
  printf("    -c TYPE,   --config=TYPE        Create a new lineakd.conf file for keyboard TYPE\n"
         "                                      (warning: old one will be overwritten!)\n");
//  printf("    -k TYPE,   --keyboard=TYPE      Specify the keyboard type (refer to -l)\n");
  printf("    -d DEVICE, --cdrom-dev=DEVICE   Specify the CDROM device\n");
  printf("    -b,        --background         Run in background\n");
  printf("    -v,        --verbose            Print verbose messages\n"
         "                                      (only visible when not running in background)\n");
  printf("    -h,        --help               Print this message\n");
#else
  printf("    -l            Show all supported keyboard types\n");
  printf("    -c TYPE       Create a new lineakd.conf file for TYPE keyboard\n"
         "                    (warning: old one will be overwritten!)\n");
//  printf("    -k TYPE       Specify the keyboard type (refer to -l)\n");
  printf("    -d DEVICE     Specify the CDROM device\n");
  printf("    -b            Run in background\n");
  printf("    -v,           Print verbose messages\n"
         "                    (only visible when not running in background)\n");
  printf("    -h            Print this message\n");
#endif
  printf("\n");
}

/* parse incoming arguments */
int parseargs(int argc, char *argv[]) {
  int myopt;
  const char *flags = "hc:bvd:l";

#ifdef HAVE_GETOPT_LONG
  int option_index = 0;
  static struct option long_options[] = {
        {"help",            0, 0, 'h'},
        {"create",          1, 0, 'c'},
        {"background",      0, 0, 'b'},
        {"verbose",         0, 0, 'v'},
//        {"keyboard",        1, 0, 'k'},
        {"cdrom-dev",       1, 0, 'd'},
        {"kbd-list",        0, 0, 'l'},
        {"verbose",         0, 0, 'v'},
        {0, 0, 0, 0}
  };

  while ((myopt = getopt_long(argc, argv, flags, long_options, &option_index)) != -1) {
#else
  while ((myopt = getopt(argc, argv, flags)) != -1) {
#endif

	/* parse each option */
    switch (myopt) {
//	      case 'k':
//              free(kbtype);
//			  kbtype = malloc((strlen(optarg)+1)*sizeof(char));
//		      strcpy(kbtype, optarg);
//		      strcat(kbtype, "\0");
//              break;
          case 'd':
              free(cdromdev);
			  cdromdev = malloc((strlen(optarg)+1)*sizeof(char));
		      strcpy(cdromdev, optarg);
		      strcat(cdromdev, "\0");
              break;
          case 'h':
              showusage(argc, argv);
              cleanexit(3);
              break;
          case 'c':
              create_new_conf(NULL,optarg);
              cleanexit(3);
              break;
          case 'b':
              runasdaemon = 1;
              break;
          case 'v':
              verbosemode = 1;
              break;
          case 'l':
              showkeyboards();
              cleanexit(3);
              break;
    }
  }
  
  /* set the program name without the directories */
  if ( strrchr( argv[0], '/' ) ) {
      programname = (char*) strrchr( argv[0], '/' ) + 1;
  } else {
      programname = argv[0];
  }

  /* do we have a valid keyboard type? */
  if (cfgSectionNameToNumber(kbtype) == -1) {
	printf("\nInvalid keyboard type: %s\nSpecify a valid keyboard in %s\nTo find out the keyboard types use: %s -l\n\n",kbtype,conffile,programname);
	cleanexit(1);
  }
  
  /* check for a single additional argument */
  if ((argc - optind) > 1) {
    fprintf( stderr, "%s: too many arguments\n", argv[0] );
	cleanexit(1);
  }

  return 1;
}

/**********************************************************************************************/
