/*
  LinEAK - Linux support for Easy Access and Internet Keyboards      
  Copyright (C) 2001,2002 Mark Smulders <Mark@PIRnet.nl>             
  http://lineak.sourceforge.net
                                                                        
  This program is free software; you can redistribute it and/or modify  
  it under the terms of the GNU General Public License as published by  
  the Free Software Foundation; either version 2, or (at your option)   
  any later version.                                                    
                                                                        
  This program is distributed in the hope that it will be useful,       
  but WITHOUT ANY WARRANTY; without even the implied warranty of        
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
  GNU General Public License for more details.                          
                                                                        
  You should have received a copy of the GNU General Public License     
  along with this program; if not, write to the                         
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,          
  Boston, MA 02111-1307, USA.                                           
*/

/* configuration file parsing (parsecfg (c) by Yuuki NINOMIYA, gm@debian.or.jp) */
#include "parsecfg.h"

/* lineakd defines */
#define DEFINITIONFILE "lineakkb.def" /* full path: CONFDIR/DEFINITIONFILE (usually /usr/local/etc/lineakd.conf) */
#define CONFIGFILE "lineakd.conf" /* resides in $HOME/.lineak/  */
#define PIDFILE "lineakd.pid" /* resides in $HOME/.lineak/  */

/* EAKey as member of key list */
typedef struct strEAkey {
  struct strEAkey *next;
  char *EAkeyname;
  int EAkeycode;
  KeySym EAkeysym;
  char *EAcommand;
} EAkey;

/* the EAK we will use.. */
typedef struct {
  int EAKnr;
  char *EAKtype;
  char *EAKname;
  EAkey *EAKeylist;
} EAKeyboard;

/* function prototypes */
void showusage(int argc, char *argv[]);
int parseargs(int argc, char *argv[]);
void showkeyboards (void);
int initEAK (void);
void execEAcommand(int keycode);
EAkey *lookupEAkey(int keycode);
void parseconffile(void);
void create_new_conf(char *homedir, char *type);
void parsedeffile(void);
int pid_file_exists(void);
void create_pid_file(void);
void delete_pid_file(void);
static void cleanexit (int status);
static void cleanup (void);
void cleanKeylist (void);
static void signalexit(int sig_num);
static void signalHUP(int sig_num);
void initializeX(void);
Display *GetDisplay(char* dpyName, int* opcodeRtrn, int* evBaseRtrn);
void commitXKBChanges(int tcode);
void initCDROM (void);
void DoEject(void);
void initSound(void);
void AdjustVolume(int value);

/* miscellaneous global variables and init */
static int runasdaemon = 0;
static int verbosemode = 0;
static int cdromstatus = 0;
static int ismuted = 0;
static int old_mastervol = 0;
static int nrretries = 0;
#define MAXRETRIES 25 /* prevents possible race condition, don't set this too low!! 25 will do */
static char *programname = NULL;
static char *deffile = NULL;
static char *conffile = NULL;
static EAKeyboard myEAK = { -1, NULL, NULL, NULL }; /* the keyboard we'll use */
Display *ourdisplay;
XkbDescPtr xkb;
char *displayname = NULL;
int xkbOpcode;
int xkbEventCode;
static int caughtXerror = 0;

/* special actions in .conf file */
#define EAK_EJECT "EAK_EJECT"
#define EAK_VOLUP "EAK_VOLUP"
#define EAK_VOLDOWN "EAK_VOLDOWN"
#define EAK_MUTE "EAK_MUTE"
#define EAK_SLEEP "EAK_SLEEP"

/* for volume settings using the mixer device */
#define MIXERDEV "/dev/mixer"
#define VOLUP_VALUE 5
#define VOLDOWN_VALUE -5
#define VOLMUTE_VALUE 0
#define MAX_VOLUME 100

/* we might need this .. oss emulation under BSD */
#if defined (__NetBSD__) || defined (__OpenBSD__)
#define SOUND_IOCTL(a,b,c)      _oss_ioctl(a,b,c)
#else
#define SOUND_IOCTL(a,b,c)      ioctl(a,b,c)
#endif

#define CONF_HEADER \
"# LinEAK - Linux support for Easy Access and Internet Keyboards\n" \
"#  Copyright (c) 2001,2002 Mark Smulders <Mark@PIRnet.nl>\n" \
"#  http://lineak.sourceforge.net\n" \
"#\n" \
"# lineakd configuration file\n" \
"#\n" \
"# example key configuration:\n" \
"# \tplay\t= \"xmms --play-pause\"\n" \
"# \teject\t= EAK_EJECT\n" \
"#\n" \
"# available special actions:\n" \
"# \tEAK_EJECT\n" \
"# \tEAK_VOLUP\n" \
"# \tEAK_VOLDOWN\n" \
"# \tEAK_MUTE\n" \
"# \tEAK_SLEEP\n" \
"#\n" \
"\n"

/* to read the data from .conf file (user configuration settings) */
/*  (the commands go into a dummy. we don't need them right now)  */
static char *kbtype=NULL, *cdromdev=NULL;
char *DUMMY_STRING;
//char *cmdPlay, *cmdPrevious, *cmdNext, *cmdStop, *cmdPause, *cmdEject;
//char *cmdInternet, *cmdMail, *cmdCalculator, *cmdFilemanager, *cmdMycomputer, *cmdMydocuments;
//char *cmdCommunities, *cmdVendorhome, *cmdSearch, *cmdLightbulb, *cmdShop, *cmdPrint, *cmdGo, *cmdFavorites;
cfgStruct confdata[] = {
  { "KeyboardType", CFG_STRING, &kbtype },
  { "CdromDevice", CFG_STRING, &cdromdev },
  { "Play", CFG_STRING, &DUMMY_STRING /*&cmdPlay*/ },
  { "Previous", CFG_STRING, &DUMMY_STRING /*&cmdPrevious*/ },
  { "Next", CFG_STRING, &DUMMY_STRING /*&cmdNext*/ },
  { "Stop", CFG_STRING, &DUMMY_STRING /*&cmdStop*/ },
  { "Pause", CFG_STRING, &DUMMY_STRING /*&cmdPause*/ },
  { "Eject", CFG_STRING, &DUMMY_STRING /*&cmdEject*/ },
  { "Internet", CFG_STRING, &DUMMY_STRING /*&cmdInternet*/ },
  { "Home", CFG_STRING, &DUMMY_STRING },
  { "WebBack", CFG_STRING, &DUMMY_STRING },
  { "WebForward", CFG_STRING, &DUMMY_STRING },
  { "WebStop", CFG_STRING, &DUMMY_STRING },
  { "WebReload", CFG_STRING, &DUMMY_STRING },
  { "Mail", CFG_STRING, &DUMMY_STRING /*&cmdMail*/ },
  { "Calculator", CFG_STRING, &DUMMY_STRING /*&cmdCalculator*/ },
  { "FileManager", CFG_STRING, &DUMMY_STRING /*&cmdFilemanager*/ },
  { "Terminal", CFG_STRING, &DUMMY_STRING },
  { "News", CFG_STRING, &DUMMY_STRING },
  { "MyComputer", CFG_STRING, &DUMMY_STRING /*&cmdMycomputer*/ },
  { "MyDocuments", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Communities", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Market", CFG_STRING, &DUMMY_STRING },
  { "Meeting", CFG_STRING, &DUMMY_STRING },
  { "VendorHome", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Search", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Lightbulb", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Shop", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Print", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Go", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Favorites", CFG_STRING, &DUMMY_STRING /*&cmdMydocuments*/ },
  { "Eject", CFG_STRING, &DUMMY_STRING },
  { "VolumeUp", CFG_STRING, &DUMMY_STRING },
  { "VolumeDown", CFG_STRING, &DUMMY_STRING },
  { "Mute", CFG_STRING, &DUMMY_STRING },
  { "Sleep", CFG_STRING, &DUMMY_STRING },
  { "ZoomIn", CFG_STRING, &DUMMY_STRING },
  { "ZoomOut", CFG_STRING, &DUMMY_STRING },
  { "ScrollUp", CFG_STRING, &DUMMY_STRING },
  { "ScrollDown", CFG_STRING, &DUMMY_STRING },
  { "Screensaver", CFG_STRING, &DUMMY_STRING },
  { "Power", CFG_STRING, &DUMMY_STRING },
  { "Wake", CFG_STRING, &DUMMY_STRING },
  { NULL, CFG_END, NULL }
};

/* to read the data from keyboard definition file (all possible keys) */
/*  (the keys go into a dummy. we don't need them right now)          */
int nrkeyboards;
char **longkbname;
cfgList **keylist;
int *DUMMY_INT;
//int *keyplay, *keyprevious, *keynext, *keystop, *keypause, *keyeject;
//int *keymail, *keycommunities, *keyvendorhome, *keyinternet, *keysearch, *keylightbulb, *keyshop;
//int *keyprint, *keyvolumedown, *keymute, *keyvolumeup;
//int *keymycomputer, *keymydocuments, *keyfilemanager, *keyfavorites, *keycalculator, *keysleep, *keygo;
static cfgStruct defdata[] = {
  { "LONGNAME", CFG_STRING, &longkbname },
  { "KEYS", CFG_STRING_LIST, &keylist },
  { "PLAY", CFG_INT, &DUMMY_INT /*&keyplay*/ },
  { "PREVIOUS", CFG_INT, &DUMMY_INT /*&keyprevious*/ },
  { "NEXT", CFG_INT, &DUMMY_INT /*&keynext*/ },
  { "STOP", CFG_INT, &DUMMY_INT /*&keystop*/ },
  { "PAUSE", CFG_INT, &DUMMY_INT /*&keypause*/ },
  { "EJECT", CFG_INT, &DUMMY_INT /*&keyeject*/ },
  { "MAIL", CFG_INT, &DUMMY_INT /*&keymail*/ },
  { "COMMUNITIES", CFG_INT, &DUMMY_INT /*&keycommunities*/ },
  { "MARKET", CFG_STRING, &DUMMY_STRING },
  { "MEETING", CFG_STRING, &DUMMY_STRING },
  { "VENDORHOME", CFG_INT, &DUMMY_INT /*&keyvendorhome*/ },
  { "INTERNET", CFG_INT, &DUMMY_INT /*&keyinternet*/ },
  { "HOME", CFG_STRING, &DUMMY_STRING },
  { "WEBBACK", CFG_STRING, &DUMMY_STRING },
  { "WEBFORWARD", CFG_STRING, &DUMMY_STRING },
  { "WEBSTOP", CFG_STRING, &DUMMY_STRING },
  { "WEBRELOAD", CFG_STRING, &DUMMY_STRING },
  { "SEARCH", CFG_INT, &DUMMY_INT /*&keysearch*/ },
  { "LIGHTBULB", CFG_INT, &DUMMY_INT /*&keylightbulb*/ },
  { "SHOP", CFG_INT, &DUMMY_INT /*&keyshop*/ },
  { "PRINT", CFG_INT, &DUMMY_INT /*&keyprint*/ },
  { "VOLUMEDOWN", CFG_INT, &DUMMY_INT /*&keyvolumedown*/ },
  { "MUTE", CFG_INT, &DUMMY_INT /*&keymute*/ },
  { "VOLUMEUP", CFG_INT, &DUMMY_INT /*&keyvolumeup*/ },
  { "MYCOMPUTER", CFG_INT, &DUMMY_INT /*&keymycomputer*/ },
  { "MYDOCUMENTS", CFG_INT, &DUMMY_INT /*&keymydocuments*/ },
  { "FILEMANAGER", CFG_INT, &DUMMY_INT /*&keyfilemanager*/ },
  { "TERMINAL", CFG_STRING, &DUMMY_STRING },
  { "NEWS", CFG_STRING, &DUMMY_STRING },
  { "FAVORITES", CFG_INT, &DUMMY_INT /*&keyfavorites*/ },
  { "CALCULATOR", CFG_INT, &DUMMY_INT /*&keycalculator*/ },
  { "SLEEP", CFG_INT, &DUMMY_INT /*&keysleep*/ },
  { "GO", CFG_INT, &DUMMY_INT /*&keygo*/ },
  { "ZoomIn", CFG_STRING, &DUMMY_STRING },
  { "ZoomOut", CFG_STRING, &DUMMY_STRING },
  { "ScrollUp", CFG_STRING, &DUMMY_STRING },
  { "ScrollDown", CFG_STRING, &DUMMY_STRING },
  { "Screensaver", CFG_STRING, &DUMMY_STRING },
  { "Power", CFG_STRING, &DUMMY_STRING },
  { "Wake", CFG_STRING, &DUMMY_STRING },
  { NULL, CFG_END, NULL }
};

  